"""
This file contains the database models
"""

from django.db.models import Model, CharField, DateTimeField, BooleanField, ForeignKey, CASCADE, IntegerField, FloatField, SET_NULL, PROTECT, ManyToManyField
from django.contrib.postgres.fields import ArrayField, JSONField
from django.contrib.postgres.indexes import GinIndex
from enum import Enum

#
# Common
#

# abstract models

class BasicCommon(Model):
    # todo: we cannot use foreign keys in the array here, so we have to keep the Tags table up to date by trigger or so.
    # todo: we could switch to a manytomany field instead?
    tags = ArrayField(CharField(max_length=128), size=8, blank=True, help_text='User-defined search keywords for object.', default=list)
    created_at = DateTimeField(auto_now_add=True, help_text='Moment of object creation.')
    updated_at = DateTimeField(auto_now=True, help_text='Moment of last object update.')

    class Meta:
        abstract = True
        indexes = [GinIndex(fields=['tags'])]


class NamedCommon(BasicCommon):
    name = CharField(max_length=128, help_text='Human-readable name of this object.', null=False) # todo: check if we want to have this primary_key=True
    description = CharField(max_length=255, help_text='A longer description of this object.')

    def __str__(self):
        return self.name

    class Meta:
        abstract = True


# todo: check if we want to generally use this flavour, i.e. make everything named addressable by name rather than int id. (This then does not allow for multiple items of same name, of course.)
class NamedCommonPK(NamedCommon):
    name = CharField(max_length=128, help_text='Human-readable name of this object.', null=False, primary_key=True)

    class Meta:
        abstract = True


# concrete models

class Tags(Model):
    # todo: figure out how to keep this in sync with tags columns (->BasicCommon)
    # todo: Or remove this altogether without keeping track of tags?
    title = CharField(max_length=128)
    description = CharField(max_length=255)


#
# I/O
#

# choices

class AbstractChoice(Model):
    """
    Abstract class for all derived 'choices' models.
    We define a 'choice' as an item that you can pick from a predefined list.
    In the derived classes, we use an enum.Enum to define such a predefined list here in code.
    All values of the enums are then put automagically into the database in the populate module, which
    is/can_be/should_be called in the last migration step to populate the database with inital values
    for our 'static choices'.

    Design decision: Django also provides the 'choices' property on fields which sort of limits the number of choices
    one can make, and which sort of does some validation. In our opinion the validation is done in the wrong place, and
    no data consistency is enforced.
    So, we decided to follow Django's own hint, see https://docs.djangoproject.com/en/2.0/ref/models/fields/#choices
    "you’re probably better off using a proper database table with a ForeignKey"

    You can find the derived AbstractChoice classes being used as ForeignKey in other models, thus enforcing data
    consistency at database level.
    """
    value = CharField(max_length=128, editable=True, null=False, blank=False, unique=True, primary_key=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.value


class Role(AbstractChoice):
    """Defines the model and predefined list of possible Role's for TaskConnectors.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        CORRELATOR = "correlator"
        BEAMFORMER = "beamformer"
        INSPECTION_PLOTS = "inspection plots"
        CALIBRATOR = "calibrator"
        TARGET = "target"
        INPUT = "input"
        OUTPUT = "output"


class Datatype(AbstractChoice):
    """Defines the model and predefined list of possible Datatype's for TaskConnectors.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        VISIBILITIES = "visibilities"
        TIME_SERIES = "time series"
        INSTRUMENT_MODEL = "instrument model"
        IMAGE = "image"
        QUALITY = "quality"


class Dataformat(AbstractChoice):
    """Defines the model and predefined list of possible Dataformat's for TaskRelationDraft and TaskRelationBlueprint.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        MEASUREMENTSET = "MeasurementSet"
        BEAMFORMED = "Beamformed"


class CopyReason(AbstractChoice):
    """Defines the model and predefined list of possible CopyReason's for SchedulingUnitDraft and TaskDraft.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        TEMPLATE = "template"
        REPEATED = "repeated"


# concrete models

class TaskConnectors(BasicCommon):
    role = ForeignKey('Role', null=False, on_delete=PROTECT)
    datatype = ForeignKey('Datatype', null=False, on_delete=PROTECT)
    dataformats = ManyToManyField('Dataformat', blank=True)
    output_of = ForeignKey("TaskTemplate", related_name='outputs', on_delete=CASCADE)
    input_of = ForeignKey("TaskTemplate", related_name='inputs', on_delete=CASCADE)


#
# Templates
#

# abstract models

class Template(NamedCommon):
    version = CharField(max_length=128, help_text='Version of this template (with respect to other templates of the same name).')
    schema = JSONField(help_text='Schema for the configurable parameters needed to use this template.')

    class Meta:
        abstract = True

# concrete models

class GeneratorTemplate(Template):
    create_function = CharField(max_length=128, help_text='Python function to call to execute the generator.')


class DefaultGeneratorTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("GeneratorTemplate", on_delete=PROTECT)


class SchedulingUnitTemplate(Template):
    pass


class DefaultSchedulingUnitTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("SchedulingUnitTemplate", on_delete=PROTECT)


class TaskTemplate(Template):
    validation_code_js = CharField(max_length=128, help_text='JavaScript code for additional (complex) validation.')


class DefaultTaskTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("TaskTemplate", on_delete=PROTECT)


class WorkRelationSelectionTemplate(Template):
    pass


class DefaultWorkRelationSelectionTemplate(BasicCommon):
    name = CharField(max_length=128, unique=True)
    template = ForeignKey("WorkRelationSelectionTemplate", on_delete=PROTECT)


#
# Instance Objects
#

class Cycle(NamedCommonPK):
    start = DateTimeField(help_text='Moment at which the cycle starts, that is, when its projects can run.')
    stop = DateTimeField(help_text='Moment at which the cycle officially ends.')
    number = IntegerField(help_text='Cycle number.')
    standard_hours = IntegerField(help_text='Number of offered hours for standard observations.')
    expert_hours  = IntegerField(help_text='Number of offered hours for expert observations.')
    filler_hours  = IntegerField(help_text='Number of offered hours for filler observations.')


class Project(NamedCommonPK):
    # cycle is protected since we have to manually decide to clean up projects with a cycle or keep them without cycle
    cycle = ForeignKey('Cycle', related_name='projects', on_delete=PROTECT, null=True, help_text='Cycle(s) to which this project belongs (NULLable).')
    priority = IntegerField(default=0, help_text='Priority of this project w.r.t. other projects. Projects can interrupt observations of lower-priority projects.')  # todo: define a value for the default priority
    can_trigger = BooleanField(default=False, help_text='True if this project is allowed to supply observation requests on the fly, possibly interrupting currently running observations (responsive telescope).')
    private_data = BooleanField(default=True, help_text='True if data of this project is sensitive. Sensitive data is not made public.')
    expert = BooleanField(default=False, help_text='Expert projects put more responsibility on the PI.')
    filler = BooleanField(default=False, help_text='Use this project to fill up idle telescope time.')

class ProjectQuotum(NamedCommonPK):
    project = ForeignKey('Project', related_name="project_quotums", on_delete=PROTECT, help_text='Project to wich this quota belongs.')  # protected to avoid accidents
    resource_quota = FloatField(help_text='Resource Quota value')
    resource_type = ForeignKey('ResourceType', related_name="resource_type", on_delete=PROTECT, help_text='Resource type.')  # protected to avoid accidents
    #lta_storage = FloatField(help_text='LTA storage offered for the project')
    #observation_hours = IntegerField(help_text='Number of offered hours for observations.')
    #processing_hours = IntegerField(help_text='Number of offered hours for  processing.')

class ResourceType(NamedCommonPK):
    resource_unit = ForeignKey('ResourceUnit', related_name="resource_types", on_delete=PROTECT, help_text='Unit of current resource.')

class ResourceUnit(NamedCommonPK):
    pass

class SchedulingSet(NamedCommon):
    generator_doc = JSONField(null=True, help_text='Parameters for the generator (NULLable).')
    generator_template = ForeignKey('GeneratorTemplate', on_delete=SET_NULL, null=True, help_text='Generator for the scheduling units in this set (NULLable).')
    generator_source = ForeignKey('SchedulingUnitDraft', on_delete=SET_NULL, null=True, help_text='Reference for the generator to an existing collection of specifications (NULLable).')
    project = ForeignKey('Project', related_name="scheduling_sets", on_delete=PROTECT, help_text='Project to which this scheduling set belongs.')  # protected to avoid accidents

class SchedulingUnitDraft(NamedCommon):
    requirements_doc = JSONField(help_text='Scheduling and/or quality requirements for this run.')
    copies = ForeignKey('SchedulingUnitDraft', related_name="copied_from", on_delete=SET_NULL, null=True, help_text='Source reference, if we are a copy (NULLable).')
    copy_reason = ForeignKey('CopyReason', null=True, on_delete=PROTECT, help_text='Reason why source was copied (NULLable).')
    generator_instance_doc = JSONField(null=True, help_text='Parameter value that generated this run draft (NULLable).')
    scheduling_set = ForeignKey('SchedulingSet', related_name='scheduling_unit_drafts', on_delete=CASCADE, help_text='Set to which this scheduling unit draft belongs.')
    requirements_template = ForeignKey('SchedulingUnitTemplate', on_delete=CASCADE, help_text='Schema used for requirements_doc.') # todo: 'schema'?


class SchedulingUnitBlueprint(NamedCommon):
    requirements_doc = JSONField(help_text='Scheduling and/or quality requirements for this scheduling unit (IMMUTABLE).')
    do_cancel = BooleanField()
    requirements_template = ForeignKey('SchedulingUnitTemplate', on_delete=CASCADE, help_text='Schema used for requirements_doc (IMMUTABLE).')
    draft = ForeignKey('SchedulingUnitDraft', related_name='related_scheduling_unit_blueprint', on_delete=CASCADE, help_text='Scheduling Unit Draft which this run instantiates.')


class TaskDraft(NamedCommon):
    specifications_doc = JSONField(help_text='Specifications for this task.')
    copies = ForeignKey('TaskDraft', related_name="copied_from", on_delete=SET_NULL, null=True, help_text='Source reference, if we are a copy (NULLable).')
    copy_reason = ForeignKey('CopyReason', on_delete=PROTECT, null=True, help_text='Reason why source was copied (NULLable).')
    scheduling_unit_draft = ForeignKey('SchedulingUnitDraft', related_name='task_drafts', on_delete=CASCADE, help_text='Scheduling Unit draft to which this task draft belongs.')
    specifications_template = ForeignKey('TaskTemplate', on_delete=CASCADE, help_text='Schema used for requirements_doc.') # todo: 'schema'?


class TaskBlueprint(NamedCommon):
    specifications_doc = JSONField(help_text='Schedulings for this task (IMMUTABLE).')
    do_cancel = BooleanField(help_text='Cancel this task.')
    specifications_template = ForeignKey('TaskTemplate', on_delete=CASCADE, help_text='Schema used for specifications_doc (IMMUTABLE).')
    draft = ForeignKey('TaskDraft', related_name='related_task_blueprint', on_delete=CASCADE, help_text='Task Draft which this task instantiates.')
    scheduling_unit_blueprint = ForeignKey('SchedulingUnitBlueprint', on_delete=CASCADE, help_text='Scheduling Unit Blueprint to which this task belongs.')


class TaskRelationDraft(BasicCommon):
    selection_doc = JSONField(help_text='Filter for selecting dataproducts from the output role.')
    dataformat = ForeignKey('Dataformat', null=False, on_delete=PROTECT, help_text='Selected data format to use. One of (MS, HDF5).')
    producer = ForeignKey('TaskDraft', related_name='produced_by', on_delete=CASCADE, help_text='Task Draft that has the output connector. NOTE: The producer does typically, but not necessarily, belong to the same Scheduling Unit (or even the same Project) as the consumer.')
    consumer = ForeignKey('TaskDraft', related_name='consumed_by', on_delete=CASCADE, help_text='Task Draft that has the input connector.')
    input = ForeignKey('TaskConnectors', related_name='inputs_task_relation_draft', on_delete=CASCADE, help_text='Input connector of consumer.')
    output = ForeignKey('TaskConnectors', related_name='outputs_task_relation_draft', on_delete=CASCADE, help_text='Output connector of producer.')
    selection_template = ForeignKey('WorkRelationSelectionTemplate', on_delete=CASCADE, help_text='Schema used for selection_doc.') # todo: 'schema'?


class TaskRelationBlueprint(BasicCommon):
    selection_doc = JSONField(help_text='Filter for selecting dataproducts from the output role.')
    dataformat = ForeignKey('Dataformat', null=False, on_delete=PROTECT, help_text='Selected data format to use.')
    producer = ForeignKey('TaskBlueprint', related_name='produced_by', on_delete=CASCADE, help_text='Task Blueprint that has the output connector.')
    consumer = ForeignKey('TaskBlueprint', related_name='consumed_by', on_delete=CASCADE, help_text='Task Blueprint that has the input connector.')
    input = ForeignKey('TaskConnectors', related_name='inputs_task_relation_blueprint', on_delete=CASCADE, help_text='Input connector of consumer.')
    output = ForeignKey('TaskConnectors', related_name='outputs_task_relation_blueprint', on_delete=CASCADE, help_text='Output connector of producer.')
    draft = ForeignKey('TaskRelationDraft', on_delete=CASCADE, related_name='related_task_relation_blueprint', help_text='Task Relation Draft which this work request instantiates.')
    selection_template = ForeignKey('WorkRelationSelectionTemplate', on_delete=CASCADE, help_text='Schema used for selection_doc.')  # todo: 'schema'?